#
# tohtml.py
#
# Implements an SAX2 ContentHandler that converts an XML document
# compliant with xmlmp 1.1 DTD to an html page.
#
# by Nick Patavalis (npat@efault.net)
#
# This is part of the xmlmp program which is free software; you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# The xmlmp program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the xmlmp program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
#

import sys
import xml.sax
import xml.sax.handler
import xmlmp
import _printer

##########################################################################

_manpage_start = """
<html>
<head>
<title>%(title)s(%(section)s)</title>
</head>

<body>

<center><table width=\"100%%\"><tr>
  <td align=\"left\">%(title)s(%(section)s)</td>
  <td align=\"center\">%(source)s</td>  
  <td align=\"right\">%(title)s(%(section)s)</td>
</tr></table></center>

<h2>NAME</h2>
<center><table width=\"90%%\"><tr><td>
%(title)s -- %(desc)s
</td></tr></table></center>
"""

_manpage_end = """

<center><table width=\"100%%\"><tr>
  <td align=\"left\">%(manual)s</td>
  <td align=\"center\">%(date)s</td>
  <td align=\"right\">%(title)s(%(section)s)</td>
</tr></table></center>

</body>
</html>
"""

_synopsis_start = """

<h2>SYNOPSIS</h2>
<center><table width=\"90%%\"><tr><td>
<ul>
"""

_synopsis_end = """
</ul>
</td></tr></table></center>
"""

_description_start = """

<h2>DESCRIPTION</h2>
<center><table width=\"90%%\"><tr><td>
"""

_description_end = """
</td></tr></table></center>
"""

_section_start = """

<h2>%(name)s</h2>
<center><table width=\"90%%\"><tr><td>
"""

_section_end = """
</td></tr></table></center>
"""

_synel_start = """
<li><b>
"""
_synel_end = """
</b></li>
"""

_synsep = """
</ul>
<hr align=\"left\" width=\"10%%\">
<ul>
"""

_dl_start = """
<dl>
"""
_dl_end = """
</dl>
"""

_ul_start = """
<ul>
"""
_ul_end = """
</ul>
"""

_ol_start = """
<ol>
"""
_ol_end = """
</ol>
"""

_li_start = """
<li>
"""
_li_end = """
</li>
"""

_dt_start = """
<dt><b>
"""
_dt_end = """
</b></dt>
"""
_dd_start = """
<dd>
"""
_dd_end = """
</dd>
"""

_p_start = """
<p>
"""
_p_end = """
</p>
"""

_cmd_start = "<b>"
_cmd_end = "</b>"

_arg_start = "<i>"
_arg_end = "</i>"

_fn_start = "<b>"
_fn_end = "</b>"

_file_start = "<b>"
_file_end = "</b>"

_manref = "<a href=\"%(href)s\"><b>%(name)s</b>(%(section)s)</a>"
_url = "<a href=\"%(href)s\">%(href)s</a>"

##########################################################################

_trdict = { "manpage" : (_manpage_start, _manpage_end, 0, 0),
            "synopsis" : (_synopsis_start, _synopsis_end, 2, -2),
            "synel" : (_synel_start, _synel_end, 2, -2),
            "synsep" : (_synsep, None, -2, 2),
            "description" : (_description_start, _description_end, 2, -2),
            "section" : (_section_start, _section_end, 2, -2),
            "dl" : (_dl_start, _dl_end, 2, -2),
            "ul" : (_ul_start, _ul_end, 2, -2),
            "ol" : (_ol_start, _ol_end, 2, -2),
            "li" : (_li_start, _li_end, 2, -2),
            "dt" : (_dt_start, _dt_end, 2, -2),
            "dd" : (_dd_start, _dd_end, 2, -2),
            "p" : (_p_start, _p_end, 2, -2),
            "cmd" : (_cmd_start, _cmd_end, 0, 0),
            "arg" : (_arg_start, _arg_end, 0, 0),
            "fn" : (_fn_start, _fn_end, 0, 0),
            "file" : (_file_start, _file_end, 0, 0),
            "manref" : (_manref, None, 0, 0),
            "url" : (_url, None, 0, 0),
            "" : (None, None, None, None) }

##########################################################################

class XmlMp2HTML (xml.sax.handler.ContentHandler):

    """ A xmlmp -> HTML ContentHandler class.

    This is a SAX2 ContentHandler class that can parse documents
    complying with the xmlmp 1.1 DTD, available at:

      http://efault.net/npat/dtds/xmlmp/xmlmp-1.1.dtd

    The output of the parser is an HTML document, visually similar to
    a unix manpage. The interface of this class is described at:

      http://python.org/doc/2.3.3/lib/content-handler-objects.html

    """

    # ATTENTION: HTML rendering is a little naive at the moment. It
    # works reasonably well, but requires some more refining to be
    # considered done.

    def __init__(self, out = sys.stdout):
        xml.sax.handler.ContentHandler.__init__(self)
        self.pr = _printer.Printer(wrap=1)

        # used to save attributes, so that they are also available at
        # the end of the elements

        self._attrs = {}

        # default values for the implied attributes of various
        # entities
        
        self._defattrs = {
            'manpage' : { u'date' : '',
                          u'source' : '',
                          u'manual' : '' },
            'manref' : { u'href' : '' },
            'url' : { u'href' : ''} }
        
    def startElement(self, name, attrs):
        # check DTD version
        if name == 'manpage' :
            version = attrs['xmlmp_ver'].split('.')
            vmaj, vmin = int(version[0]), int(version[1])
            if vmaj != xmlmp.vmaj or xmlmp.vmin < vmin :
                raise xmlmp.VersionError, \
                      "Bad xmlmp version: %d.%d. Must be %d.x (x <= %d)" \
                      % (vmaj, vmin, xmlmp.vmaj, xmlmp.vmin)
        self._attrs[name] = self._defattrs.get(name, {}).copy()
        for k in attrs.keys() :
            self._attrs[name][k] = attrs[k]
        self._attrs[name]['__name__'] = name
        if _trdict.has_key(name) and _trdict[name][0] :
            if _trdict[name][1] is None :
                # this is an empty element. handle alignment
                # specially
                self.pr.indent(_trdict[name][2])
                self.pr.pr_pre(_trdict[name][0] % self._attrs[name])
                self.pr.indent(_trdict[name][3])                
            else :
                # this is a non-empty element
                self.pr.pr_pre(_trdict[name][0] % self._attrs[name])
                self.pr.indent(_trdict[name][2])

    def endElement(self, name):
        if _trdict.has_key(name) and _trdict[name][1] :
            self.pr.indent(_trdict[name][3])            
            self.pr.pr_pre(_trdict[name][1] % self._attrs[name])

    def characters(self, content):
        self.pr.pr(content)

##########################################################################

def make_parser (out = sys.stdout):
    
    """ Create an xmlmp2html reader

    Creates and returns an SAX2 XMLReader object. The reader returned
    is already connected to the appropriate handlers, and is
    ready-and-set to parse xmlmp documents and output HTML page. For a
    description of the interface of the returned object see:

      http://python.org/doc/2.3.3/lib/xmlreader-objects.html

    The only method you will most-probably need is 'parse' which takes
    one argument, being a file object or a file-name corresponding to
    the document you wish to parse.
    
    """
    
    parser = xml.sax.make_parser()
##    parser.setEntityResolver(MyResolver())
    parser.setContentHandler(XmlMp2HTML(out = out))
    return parser
