#
# toman.py
#
# Implements an SAX2 ContentHandler that converts an XML document
# compliant with xmlmp 1.1 DTD to a unix manpage (groff script).
#
# by Nick Patavalis (npat@efault.net)
#
# This is part of the xmlmp program which is free software; you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# The xmlmp program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the xmlmp program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
#

import sys
import xml.sax
import xml.sax.handler
import xmlmp
import _printer

_manpage_start = """
.TH \"%(title)s\" \"%(section)s\" \"%(date)s\" \"%(source)s\" \"%(manual)s\"
.SH NAME
%(title)s \- %(desc)s
"""

_synopsis_start = """
.SH SYNOPSIS
"""

_description_start = """
.SH DESCRIPTION
"""

_section_start = """
.SH %(name)s
"""

class XmlMp2Man (xml.sax.handler.ContentHandler):

    """ A xmlmp -> man (groff) ContentHandler class.

    This is a SAX2 ContentHandler class that can parse documents
    complying with the xmlmp 1.1 DTD, available at:

      http://efault.net/npat/dtds/xmlmp/xmlmp-1.1.dtd

    The output of the parser is a unix man-page (groff script), that
    can be viewed by man(1), or processed by groff(1). The interface
    of this class is described at:

      http://python.org/doc/2.3.3/lib/content-handler-objects.html

    """

    def __init__(self, out = sys.stdout):
        xml.sax.handler.ContentHandler.__init__(self)
        self.pr = _printer.Printer(wrap=0)
        
        # roff cannot handle nested font changes, so we have to
        # maintain a stack with the previously-current fonts
        # ourselves.  actually the manpage dtd does not allow
        # font-changing elements to nest, with the (marginal)
        # exception of <arg> inside <synel> (<synel> implies '.B
        # '). we could therefore possibly get away without a font
        # stack, but I want to have things prepared in case nestable
        # font-changing entities are intoduced in the future. In this
        # implementation the entities <arg>, <cmd>, <fn>, <file>,
        # <url>, and <manref> are implemented as if they could nest
        # within each other.
        
        self._fstack = []

        # _cfont keeps the current font, in the form of a mandoc
        # font-change command. '.B ' for bold, '.I ' for italic, and
        # '' for normal (roman). _cfont_apply is a flag used to delay
        # the application of the font-change commands until there is
        # some text that uses the new font. when the font is to be
        # changed _cfont is pushed in the stack, _cfont is updated
        # with the new font, and _cfont_apply is set. the actuall
        # font-change command is issued only when there is text to
        # write. this way unnecessary font-changes are avoided. for
        # example the input "<arg>foo<cmd>bar</cmd></arg> foobar",
        # without delayed application would produce:
        #
        #     .I foo
        #     .B bar
        #     .I
        #     foobar
        #
        # while with the delayed-application scheme described:
        #
        #     .I foo
        #     .B bar
        #     foobar
        
        self._cfont = ''
        self._cfont_apply = 0

        # _indent_level and _rs_level are used to keep track of
        # multi-level lists (i.e. nested <ul>s, <ol>s, and
        # <dl>s). _indent_level reflects the current nesting level. it
        # is 0 in the toplevel, 1 in the fist list, 2 in the second
        # nested list, and so on. _rs_level reflects the current
        # left-margin offset (.RS level) which may lag behing the
        # current indent level. _indent_level is incremented every
        # time a new list is introduced and decremented every time a
        # list is closed. _rs_level is incremented every time an
        # ".RS" command is issued and decremented every time an '.RE'
        # command is issued. the tricky part is figuring out when to
        # issue '.RS'es and '.RE's. '.RS'es are issued when a new list
        # (say, of level 'n') is started and the _rs_level is lower
        # than the _indent_level (indicating that we are working with
        # the "prevailing indent" of the n-1 list). In this case an
        # '.RS' is issued in order to push the left-margin to the same
        # offset as the prevailing indent. The left-margin must be
        # pulled-back when the *item* of *n-1* list that contained the
        # n-list is closed. Schematically it goes like this ('i' is
        # the _indent_level and 'r' is the _rs_level):
        #
        #   i r
        #   0 0  texttexttexttext
        #   1 0  * texttexttextte
        #   1 0    texttexttextte
        #   1 0  * texttexttextte
        #   2 1    * texttexttext <- RS: <ul> met, and i > r
        #   2 1      texttexttext
        #   2 1    * texttexttext
        #   1 1    texttexttextte
        #   1 1    texttexttextte <- RE: </li> met, and i <= r
        #   1 0  * texttexttextte 
        #   1 0    texttexttextte
        #   0 0  texttexttexttext
        
        self._indent_level = 0
        self._rs_level = 0

        # _par_pending is used to delay paragraph-change commands
        # ('.PP's, '\n\n's, etc) in a manner similar to delayed font
        # application. When the end of a paragraph element (</p>) is
        # met, instead of emmiting the paragraph-change command on the
        # spot, we save it in _par_pending, and only issue it when
        # the start of the next element is met. this allows for a
        # pending par-change command to be "canceled" if it is
        # followed by a command that implies the paragraph-change
        # operation. For example the input
        #
        #   "<p>foo bar</p></li><li><p>foobar</p>",
        #
        # without delayed par-change application would produce:
        #
        #   foo bar
        #   \n
        #   .TP 3
        #   foobar
        #
        # the paragraph-change followed py the '.TP' command, has the
        # undesirable effect of producing too much vertical space.
        # With delayed par-change application this becomes:
        #
        #   foo bar
        #   .TP 3
        #   foobar
        #
        # Here the the start of the <li> element cancels the
        # paragraph-change command.
        
        self._par_pending = ''

        # default values for the implied attributes of various
        # entities
        
        self._defattrs = {
            'manpage' : { u'date' : '',
                          u'source' : '',
                          u'manual' : '' } }

    def startElement(self, name, attrs):

        # handle any delayed paragraph-change command
        if self._par_pending :
            if ( name == 'ul'
                 or name == 'ol'
                 or name == 'dl'
                 or name == 'li'
                 or name == 'dt' ) :
                # cancel the command. these elements produce
                #  commands that imply the paragraph-change
                pass
            else :
                # issue delayed paragraph-change
                self.pr.pr_pre(self._par_pending)
            self._par_pending = ''

        if name == 'manpage' :
            # check DTD version
            version = attrs['xmlmp_ver'].split('.')
            vmaj, vmin = int(version[0]), int(version[1])
            if vmaj != xmlmp.vmaj or xmlmp.vmin < vmin :
                raise xmlmp.VersionError, \
                      "Bad xmlmp version: %d.%d. Must be %d.x (x <= %d)" \
                       % (vmaj, vmin, xmlmp.vmaj, xmlmp.vmin)
            # fill-in defaults and process the element
            my_attrs = self._defattrs.get(name, {}).copy()
            for k in attrs.keys() :
                my_attrs[k] = attrs[k]            
            self.pr.pr_pre(_manpage_start % my_attrs)
        elif name == 'synopsis' :
            self.pr.pr_pre(_synopsis_start)
        elif name == 'description' :
            self.pr.pr_pre(_description_start)            
        elif name == 'section' :
            self.pr.pr_pre(_section_start % attrs)

        elif name == 'manref' :
            self._fstack.append(self._cfont)
            self._cfont = ".B "
            self._cfont_apply = 1
            self.characters("%s(%s)" % (attrs['name'], attrs['section']))
            self._cfont = self._fstack.pop(-1)
            self._cfont_apply = 1
##            self.pr.pr_pre("\n.B %s(%s)\n"
##                           % (attrs['name'], attrs['section']))
        elif name == 'url' :
            self._fstack.append(self._cfont)
            self._cfont = ".B "
            self._cfont_apply = 1
            self.characters("%s" % attrs['href'])
            self._cfont = self._fstack.pop(-1)
            self._cfont_apply = 1            
##            self.pr.pr_pre("\n.B %s\n" % attrs['href'])
                            
        elif name == 'synel' :
            self._cfont = ".B "
            self._cfont_apply = 1
        elif name == 'synsep':
            self.pr.pr_pre('\n.sp\n')
            
        elif name == 'arg' :
            self._fstack.append(self._cfont)
            self._cfont = ".I "
            self._cfont_apply = 1
        elif name == 'cmd' :
            self._fstack.append(self._cfont)
            self._cfont = ".B "
            self._cfont_apply = 1
        elif name == 'fn' :
            self._fstack.append(self._cfont)
            self._cfont = ".B "
            self._cfont_apply = 1
        elif name == 'file' :
            self._fstack.append(self._cfont)
            self._cfont = ".I "
            self._cfont_apply = 1

        elif ( name == 'dl'
               or name == 'ul'
               or name == 'ol' ) :
            if self._indent_level > self._rs_level :
                self.pr.pr_pre("\n.RS 3\n")
                self._rs_level += 1
            self._indent_level += 1
        elif name == 'dt' :
            self.pr.pr_pre("\n.TP 3\n")
            self._cfont = ".B "
            self._cfont_apply = 1
        elif name == 'dd' :
            self.pr.pr_pre("\n")
            self._cfont = ""
            self._cfont_apply = 1
        elif name == 'li' :
            self.pr.pr_pre("\n.IP \(em 3\n")

    def endElement(self, name):

        if ( name == 'mangape'
             or name == 'synopsis'
             or name == 'description'             
             or name == 'section' ) :
            self.pr.pr_pre('\n')

        elif name == 'synel' :
            self.pr.pr_pre('\n.br\n')
            self._cfont = ''
            self._cfont_apply = 0
            
        elif ( name == 'arg'
               or name == 'cmd'
               or name == 'fn'
               or name == 'file' ) :
            self._cfont = self._fstack.pop(-1)
            self._cfont_apply = 1
            
        elif name == 'p' :
            # the '.br' and '.sp' are really ugly, but I cannot use
            # '\n\n' since it leaves larger vertical space than '.PP'.
            # Is there any other par-change command, that does *not*
            # reset the "prevailing indent"?
            self._par_pending = '\n.br\n.sp 0.6v\n'

        elif ( name == 'li'
               or name == 'dd' ):
            if self._indent_level <= self._rs_level :
                self.pr.pr_pre('\n.RE\n')
                self._rs_level -= 1

        elif ( name == 'dl'
               or name == 'ul'
               or name == 'ol' ) :
            self._indent_level -= 1
            self._par_pending = '\n.PP\n'

    def characters(self, content):
        if self._cfont_apply and content.strip() :
            self.pr.pr_pre('\n' + self._cfont)
            self._cfont_apply = 0
        self.pr.pr(content)

def make_parser (out = sys.stdout):
    
    """ Create an xmlmp2man reader

    Creates and returns an SAX2 XMLReader object. The reader returned
    is already connected to the appropriate handlers, and is
    ready-and-set to parse xmlmp documents and output unix manual
    pages (groff scripts). For a description of the interface of the
    returned object see:

      http://python.org/doc/2.3.3/lib/xmlreader-objects.html

    The only method you will most-probably need is 'parse' which takes
    one argument, being a file object or a file-name corresponding to
    the document you wish to parse.
    
    """
    
    parser = xml.sax.make_parser()
##    parser.setEntityResolver(MyResolver())
    parser.setContentHandler(XmlMp2Man(out = out))
    return parser
